//===================================================================
// Программатор Atmel AT89C2051 Москва 2023 FQ=2MHz Internal
//===================================================================
#define RST_L   PORTD &= ~(1<<7)
#define RST_H   PORTD |= 1<<7
#define XT1_L   PORTD &= ~(1<<3)
#define XT1_H   PORTD |= 1<<3
#define V12_ON  PORTD &= ~(1<<6)
#define V12_OFF PORTD |= 1<<6
#define VCC_ON  PORTD &= ~(1<<5)
#define VCC_OFF PORTD |= 1<<5
#define P32_L   PORTD &= ~(1<<4)
#define P32_H   PORTD |= 1<<4
#define P33_L   PORTC &= ~(1<<0)
#define P33_H   PORTC |= 1<<0
#define P34_L   PORTC &= ~(1<<1)
#define P34_H   PORTC |= 1<<1
#define P35_L   PORTC &= ~(1<<2)
#define P35_H   PORTC |= 1<<2
#define P37_L   PORTC &= ~(1<<3)
#define P37_H   PORTC |= 1<<3
#define LED_ON  PORTC &= ~(1<<4)
#define LED_OFF PORTC |= 1<<4
#define RST 7
#define V12 6
#define VCC 5
#define P32 4
#define XT1 3
#define P33 A0
#define P34 A1
#define P35 A2
#define P37 A3
#define LED A4
#define OK '+'
#define D10mS  delayMicroseconds(10000);
#define D2mS  delayMicroseconds(2000);
#define D1mS  delayMicroseconds(1000);
#define D10uS delayMicroseconds(10);

const int pSize=18;
uint8_t buffer[pSize];
int count,i,n;
int Vpp,Vcc;
float U,FS,FS1;
unsigned long timind;

//=================================================================== 
void setup() {
  Serial.setTimeout(1000); //настраиваем тай-маут для порта
  Serial.begin(4800); // инициализация последовательного порта
  analogReference(EXTERNAL);
  pinMode(A5, INPUT);             // Вход измерения напряжения
  DDRB=0x00; // Вход
  pinMode(RST,OUTPUT);
  pinMode(V12,OUTPUT);
  pinMode(VCC,OUTPUT);
  pinMode(P32,OUTPUT);
  pinMode(P33,OUTPUT);
  pinMode(P34,OUTPUT);
  pinMode(P35,OUTPUT);
  pinMode(P37,OUTPUT);
  pinMode(XT1,OUTPUT);
  pinMode(LED,OUTPUT);
  V12_OFF; VCC_OFF; LED_OFF;
}
//=================================================================== 
void loop() {
  if(Serial.available()) {
    count=Serial.readBytes(buffer,2);
    if(count==2) {    // 
      if(buffer[0]=='G') {
		    switch(buffer[1]) {
		    case 'W':	WriteFlash();	    break; //
			  case 'E':	ChipErase();	    break; //
        case 'S':	ReadSignature();	break; //
        case 'R':	ReadFlash();	    break; //
        case 'V':	ReadVPP();	      break; //
        case 'C':	Serial.print(OK); break; //
			  default:	break;
			  }
		  }
    }
  }
}
//===================================================================
void WriteFlash(){
  boolean f=true; 
  uint8_t n=0; 
  uint8_t i; 
  Serial.print(OK);
  DDRB=0xFF; // Выход
  VCC_ON; RST_L; XT1_L; D10uS; RST_H; P32_H;
  P33_L; P34_H; P35_H; P37_H; D10uS; LED_ON;
  while (f) { // повторять до команды конец(любая кроме +)
    if(Serial.available()) {
      Serial.readBytes(buffer,1);
      if(buffer[0]=='+') { // если + то прочитать 16 байт
        while(Serial.available()<16) {};
        n=Serial.readBytes(buffer,16);
        V12_ON; // Включаем 12в
        for(i=0;i<n;i++){ // программируем 16 байт из buffer[i]
          PORTB=buffer[i]; //
          D10uS; P32_L; D10uS; P32_H; // Импульс программирования
          D2mS; // Ожидаем запись байта
          XT1_H; XT1_L; D10uS; // Меняем адрес
        }
        V12_OFF; // Выключаем 12в
        Serial.print(OK);
      }
      else f=false;
    }
  }
  PORTB=0x00; //  
  XT1_L; RST_L; P32_L; VCC_OFF; LED_OFF;
}
//===================================================================
void ReadSignature(){
  Serial.print(OK);
  DDRB=0x00; // Вход
  PORTB=0xFF; // INPUT_PULLUP
  VCC_ON; RST_L; XT1_L;  D10uS; RST_H; P32_H;
  P33_L; P34_L; P35_L; P37_L; D10uS; LED_ON;
  if(PINB<16) Serial.print('0');
  Serial.print(PINB,HEX);
  XT1_H; XT1_L; D10uS;
  if(PINB<16) Serial.print('0');
  Serial.print(PINB,HEX);
  PORTB=0x00; //
  XT1_L; RST_L; P32_L; VCC_OFF; LED_OFF;
}
//===================================================================
void ChipErase(){
  Serial.print(OK);
  VCC_ON; RST_L; XT1_L; D10uS; RST_H; P32_H; 
  P33_H; P34_L; P35_L; P37_L; D10uS; LED_ON;
  V12_ON; P32_L; D10mS P32_H; D10uS; V12_OFF; D10uS; 
  XT1_L; RST_L; P32_L; VCC_OFF; LED_OFF;
}
//===================================================================
void ReadFlash(){
  boolean f=true; 
  uint8_t n=0; 
  uint8_t i; 
  Serial.print(OK);
  DDRB=0x00; // Вход
  PORTB=0xFF; // INPUT_PULLUP
  VCC_ON; RST_L; XT1_L; D10uS; RST_H; P32_H;
  P33_L; P34_L; P35_H; P37_H; D10uS; LED_ON;
  while (f) {
    if(Serial.available()) {
      Serial.readBytes(buffer,1);
      if(buffer[0]=='+') {
        D1mS;
        for(i=0;i<16;i++){
          if(PINB<16) Serial.print('0'); 
          Serial.print(PINB,HEX);
          XT1_H; XT1_L; D10uS;
        }
       }
      else f=false;
    }
  }
  PORTB=0x00; //  
  XT1_L; RST_L; P32_L; VCC_OFF; LED_OFF;
}
//===================================================================
void ReadVPP(){
  boolean f=true; 
  V12_ON; RST_H; LED_ON;
    D10uS;
    for(i=0;i<10;i++){
      Vcc=analogRead(A5);
      U=Vcc*0.01707;
      FS = FiltrU(U);
    }
    V12_OFF;
    D10uS;
    for(i=0;i<10;i++){
      Vcc=analogRead(A5);
      U=Vcc*0.01707;
      FS1 = FiltrU(U);
    }
    V12_ON;
    Serial.print(OK);
    Serial.print(FS);
    Serial.print(FS1);
  while (f) {
    if(Serial.available()) {
      Serial.readBytes(buffer,1);
      if(buffer[0]=='+') {
        f=false;
      }  
    }
  }
  V12_OFF; RST_L; LED_OFF;
}
//=Фильтр бегущее среднее с адаптивным коэффициентом=================
float FiltrU(float newVal) {
  static float filValU = 0;
  float k;
  if (abs(newVal - filValU) > 0.1) k = 0.95;
  else k = 0.05;
  filValU += (newVal - filValU) * k;
  return filValU;
}
